# -*- coding: utf-8 -*-
"""
Tests for pyTibber
"""
import asyncio
import os
import sys
import time
import unittest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../')))
import aiohttp
import tibber


class TestTibber(unittest.TestCase):
    """
    Tests Tibber
    """

    def setUp(self):     # pylint: disable=invalid-name
        """ things to be run when tests are started. """
        self.tibber = tibber.Tibber()
        self.tibber.sync_update_info()

    def tearDown(self):  # pylint: disable=invalid-name
        """ Stop stuff we started. """
        self.tibber.sync_close_connection()

    def test_tibber(self):
        self.assertEqual(self.tibber.name, 'Arya Stark')
        self.assertEqual(len(self.tibber.get_homes()), 1)

    def test_invalid_home(self):
        home = self.tibber.get_home("INVALID_KEY")
        self.assertEqual(home, None)

    def test_home(self):
        home = self.tibber.get_homes()[0]
        home.sync_update_info()
        self.assertEqual(home.home_id, 'c70dcbe5-4485-4821-933d-a8a86452737b')
        self.assertEqual(home.address1, 'Kungsgatan 8')
        self.assertEqual(home.country, 'SE')
        self.assertEqual(home.price_unit, 'SEK/kWh')
        self.assertTrue(home.has_real_time_consumption)

        self.assertEqual(home.current_price_total, None)
        self.assertEqual(home.price_total, {})
        self.assertEqual(home.current_price_info, {})

        home.sync_update_current_price_info()
        self.assertTrue(home.current_price_total > 0)
        self.assertTrue(isinstance(home.current_price_info.get('energy'), (float, int)))
        self.assertTrue(isinstance(home.current_price_info.get('startsAt'), str))
        self.assertTrue(isinstance(home.current_price_info.get('tax'), (float, int)))
        self.assertTrue(isinstance(home.current_price_info.get('total'), (float, int)))

        home.sync_update_price_info()
        for key in home.price_total.keys():
            self.assertTrue(isinstance(key, str))
            self.assertTrue(isinstance(home.price_total[key], (float, int)))

    def test_update_info(self):
        self.assertEqual(len(self.tibber.get_homes()), 1)
        self.tibber.sync_update_info()
        self.assertEqual(len(self.tibber.get_homes()), 1)


class TestTibberWebsession(unittest.TestCase):
    """
    Tests Tibber
    """

    def setUp(self):     # pylint: disable=invalid-name
        """ things to be run when tests are started. """
        async def _create_session():
            return aiohttp.ClientSession()
        loop = asyncio.get_event_loop()
        self.websession = loop.run_until_complete(_create_session())
        self.tibber = tibber.Tibber(websession=self.websession)
        self.tibber.sync_update_info()

    def tearDown(self):  # pylint: disable=invalid-name
        """ Stop stuff we started. """
        self.tibber.sync_close_connection()

    def test_tibber(self):
        self.assertEqual(self.tibber.name, 'Arya Stark')
        self.assertEqual(len(self.tibber.get_homes()), 1)

        home = self.tibber.get_homes()[0]
        self.tibber.sync_close_connection()
        self.assertRaises(RuntimeError, self.tibber.sync_update_info)

        self.assertRaises(RuntimeError, home.sync_update_info)
        self.assertEqual(home.home_id, 'c70dcbe5-4485-4821-933d-a8a86452737b')
        self.assertEqual(home.address1, '')
        self.assertFalse(home.has_real_time_consumption)
        self.assertEqual(home.country, '')
        self.assertEqual(home.price_unit, ' ')


class TestTibberInvalidToken(unittest.TestCase):
    """
    Tests Tibber
    """

    def setUp(self):     # pylint: disable=invalid-name
        """ things to be run when tests are started. """
        self.tibber = tibber.Tibber(access_token='INVALID_TOKEN')        
        self.assertRaises(tibber.InvalidLogin, self.tibber.sync_update_info)

    def tearDown(self):  # pylint: disable=invalid-name
        """ Stop stuff we started. """
        self.tibber.sync_close_connection()

    def test_tibber(self):
        self.assertEqual(self.tibber.name, None)
        self.assertEqual(len(self.tibber.get_homes()), 0)

        async def run():
            self.assertFalse(await self.tibber.send_notification("Test tittle", "message"))
        loop = asyncio.get_event_loop()
        task = loop.create_task(run())
        loop.run_until_complete(task)


class TestTibberPrivateToken(unittest.TestCase):
    """
    Tests Tibber
    """

    def setUp(self):     # pylint: disable=invalid-name
        """ things to be run when tests are started. """
        self.tibber = tibber.Tibber(access_token='d11a43897efa4cf478afd659d6c8b7117da9e33b38232fd454b0e9f28af98012')
        self.tibber.sync_update_info()

    def tearDown(self):  # pylint: disable=invalid-name
        """ Stop stuff we started. """
        self.tibber.sync_close_connection()

    def test_tibber(self):
        self.assertEqual(self.tibber.name, 'Daniel Høyer')
        self.assertEqual(len(self.tibber.get_homes()), 0)
        self.assertEqual(len(self.tibber.get_homes(only_active=False)), 0)

    def test_invalid_home(self):
        home = self.tibber.get_home("INVALID_KEY")
        self.assertEqual(home, None)


if __name__ == '__main__':
    unittest.main()



"""Library to handle connection with Tibber API."""
import asyncio
import logging

import aiohttp
import async_timeout
from gql import gql
from graphql.language.printer import print_ast

from .subscription_manager import SubscriptionManager

DEFAULT_TIMEOUT = 10
DEMO_TOKEN = 'd1007ead2dc84a2b82f0de19451c5fb22112f7ae11d19bf2bedb224a003ff74a'
API_ENDPOINT = 'https://api.tibber.com/v1-beta/gql'
SUB_ENDPOINT = 'wss://api.tibber.com/v1-beta/gql/subscriptions'

_LOGGER = logging.getLogger(__name__)


class Tibber:
    """Class to comunicate with the Tibber api."""
    # pylint: disable=too-many-instance-attributes

    def __init__(self, access_token=DEMO_TOKEN,
                 timeout=DEFAULT_TIMEOUT,
                 websession=None):
        """Initialize the Tibber connection."""
        if websession is None:
            async def _create_session():
                return aiohttp.ClientSession()
            loop = asyncio.get_event_loop()
            self.websession = loop.run_until_complete(_create_session())
        else:
            self.websession = websession
        self._timeout = timeout
        self._access_token = access_token
        self._name = None
        self._home_ids = []
        self._all_home_ids = []
        self._homes = {}
        self.sub_manager = None

    async def close_connection(self):
        """Close the Tibber connection."""
        await self.websession.close()

    def sync_close_connection(self):
        """Close the Tibber connection."""
        loop = asyncio.get_event_loop()
        task = loop.create_task(self.close_connection())
        loop.run_until_complete(task)

    async def rt_connect(self, loop):
        """Start subscription manager for real time data."""
        if self.sub_manager is not None:
            return
        self.sub_manager = SubscriptionManager(loop,
                                               "token={}".format(self._access_token),
                                               SUB_ENDPOINT)
        self.sub_manager.start()

    async def rt_disconnect(self):
        """Stop subscription manager."""
        if self.sub_manager is None:
            return
        await self.sub_manager.stop()

    async def execute(self, document, variable_values=None):
        """Execute gql."""
        res = await self._execute(document, variable_values)
        if res is None:
            return None
        return res.get('data')

    async def _execute(self, document, variable_values=None, retry=2):
        """Execute gql."""
        query_str = print_ast(document)
        payload = {
            'query': query_str,
            'variables': variable_values or {}
        }

        post_args = {
            'headers': {'Authorization': 'Bearer ' + self._access_token},
            'data': payload
        }

        try:
            with async_timeout.timeout(self._timeout):
                resp = await self.websession.post(API_ENDPOINT,
                                                  **post_args)
            if resp.status != 200:
                _LOGGER.error("Error connecting to Tibber, resp code: %s", resp.status)
                return None
            result = await resp.json()
        except aiohttp.ClientError as err:
            _LOGGER.error("Error connecting to Tibber: %s ", err, exc_info=True)
            if retry > 0:
                return await self._execute(document, variable_values, retry-1)
            raise
        except asyncio.TimeoutError as err:
            _LOGGER.error("Timed out when connecting to Tibber: %s ", err, exc_info=True)
            if retry > 0:
                return await self._execute(document, variable_values, retry-1)
            raise
        errors = result.get('errors')
        if errors:
            _LOGGER.error('Received non-compatible response %s', errors)
        return result

    def sync_update_info(self, *_):
        """Update home info."""
        loop = asyncio.get_event_loop()
        task = loop.create_task(self.update_info())
        loop.run_until_complete(task)

    async def update_info(self, *_):
        """Update home info async."""
        query = gql('''
        {
          viewer {
            name
            homes {
              subscriptions {
                status
              }
              id
            }
          }
        }
        ''')

        res = await self._execute(query)
        if res is None:
            return
        errors = res.get('errors', [])
        if errors:
            msg = errors[0].get('message', 'failed to login')
            _LOGGER.error(msg)
            raise InvalidLogin(msg)

        data = res.get('data')
        if not data:
            return
        viewer = data.get('viewer')
        if not viewer:
            return
        self._name = viewer.get('name')
        homes = viewer.get('homes', [])
        self._home_ids = []
        for _home in homes:
            home_id = _home.get('id')
            subs = _home.get('subscriptions', [{}])
            self._all_home_ids += [home_id]
            if not subs:
                continue
            status = subs[0].get('status', 'ended').lower()
            if not home_id or not status == 'running':
                continue
            self._home_ids += [home_id]

    @property
    def name(self):
        """Return name of user."""
        return self._name

    @property
    def home_ids(self):
        """Return list of home ids."""
        return self.get_home_ids(only_active=True)

    def get_home_ids(self, only_active=True):
        """Return list of home ids."""
        if only_active:
            return self._home_ids
        return self._all_home_ids

    def get_homes(self, only_active=True):
        """Return list of Tibber homes."""
        return [self.get_home(home_id) for home_id in self.get_home_ids(only_active)]

    def get_home(self, home_id):
        """Retun an instance of TibberHome for given home id."""
        if home_id not in self._all_home_ids:
            _LOGGER.error("Could not find any Tibber home with id: %s",
                          home_id)
            return None
        if home_id not in self._homes.keys():
            self._homes[home_id] = TibberHome(home_id, self)
        return self._homes[home_id]

    async def send_notification(self, title, message):
        """Send notification."""
        query = gql('''
        mutation{
          sendPushNotification(input: {
            title: "%s",
            message: "%s",
          }){
            successful
            pushedToNumberOfDevices
          }
        }
        ''' % (title, message))

        res = await self.execute(query)
        if not res:
            return False
        noti = res.get("sendPushNotification", {})
        successful = noti.get('successful', False)
        pushed_to_number_of_devices = noti.get('pushedToNumberOfDevices', 0)
        _LOGGER.debug("send_notification: status %s, send to %s devices",
                      successful, pushed_to_number_of_devices)
        return successful


class TibberHome:
    """Instance of Tibber home."""
    # pylint: disable=too-many-instance-attributes

    def __init__(self, home_id, tibber_control):
        """Initialize the Tibber home class."""
        self._tibber_control = tibber_control
        self._home_id = home_id
        self._current_price_total = None
        self._current_price_info = {}
        self._price_info = {}
        self.sub_manager = None
        self.info = {}
        self._subscription_id = None

    def sync_update_info(self):
        """Update current price info."""
        loop = asyncio.get_event_loop()
        task = loop.create_task(self.update_info())
        loop.run_until_complete(task)

    async def update_info(self):
        """Update current price info async."""
        query = gql('''
        {
          viewer {
            home(id: "%s") {
              appNickname
              features {
                  realTimeConsumptionEnabled
                }
              address {
                address1
                address2
                address3
                city
                postalCode
                country
                latitude
                longitude
              }
              meteringPointData {
                consumptionEan
                energyTaxType
                estimatedAnnualConsumption
                gridCompany
                productionEan
                vatType
              }
              owner {
                name
                isCompany
                language
                contactInfo {
                  email
                  mobile
                }
              }
              timeZone
              subscriptions {
                id
                status
                validFrom
                validTo
                statusReason
              }
             currentSubscription {
                    priceInfo {
                      current {
                        currency
                      }
                    }
                  }
                }
              }
            }
        ''' % self._home_id)
        self.info = await self._tibber_control.execute(query)

    def sync_update_current_price_info(self):
        """Update current price info."""
        loop = asyncio.get_event_loop()
        task = loop.create_task(self.update_current_price_info())
        loop.run_until_complete(task)

    async def update_current_price_info(self):
        """Update current price info async."""
        query = gql('''
        {
          viewer {
            home(id: "%s") {
              currentSubscription {
                priceInfo {
                  current {
                    energy
                    tax
                    total
                    startsAt
                  }
                }
              }
            }
          }
        }
        ''' % self.home_id)
        price_info_temp = await self._tibber_control.execute(query)
        if not price_info_temp:
            _LOGGER.error("Could not find current price info.")
            return
        try:
            home = price_info_temp['viewer']['home']
            current_subscription = home['currentSubscription']
            price_info = current_subscription['priceInfo']['current']
        except (KeyError, TypeError):
            _LOGGER.error("Could not find current price info.")
            return
        if price_info:
            self._current_price_info = price_info

    def sync_update_price_info(self):
        """Update current price info."""
        loop = asyncio.get_event_loop()
        task = loop.create_task(self.update_price_info())
        loop.run_until_complete(task)

    async def update_price_info(self):
        """Update price info async."""
        query = gql('''
        {
          viewer {
            home(id: "%s") {
              currentSubscription {
                priceInfo {
                  current {
                    energy
                    tax
                    total
                    startsAt
                  }
                  today {
                    total
                    startsAt
                  }
                  tomorrow {
                    total
                    startsAt
                  }
                }
              }
            }
          }
        }
        ''' % self.home_id)
        price_info_temp = await self._tibber_control.execute(query)
        if not price_info_temp:
            _LOGGER.error("Could not find price info.")
            return
        self._price_info = {}
        for key in ['current', 'today', 'tomorrow']:
            try:
                home = price_info_temp['viewer']['home']
                current_subscription = home['currentSubscription']
                price_info = current_subscription['priceInfo'][key]
            except (KeyError, TypeError):
                _LOGGER.error("Could not find price info for %s.", key)
                continue
            if key == 'current':
                self._current_price_info = price_info
                continue
            for data in price_info:
                self._price_info[data.get('startsAt')] = data.get('total')

    @property
    def current_price_total(self):
        """Get current price total."""
        if not self._current_price_info:
            return None
        return self._current_price_info.get('total')

    @property
    def current_price_info(self):
        """Get current price info."""
        return self._current_price_info

    @property
    def price_total(self):
        """Get dictionary with price total, key is date-time."""
        return self._price_info

    @property
    def home_id(self):
        """Return home id."""
        return self._home_id

    @property
    def has_real_time_consumption(self):
        """Return home id."""
        try:
            return self.info['viewer']['home']['features']['realTimeConsumptionEnabled']
        except (KeyError, TypeError):
            return False

    @property
    def address1(self):
        """Return the home adress1."""
        try:
            return self.info['viewer']['home']['address']['address1']
        except (KeyError, TypeError):
            _LOGGER.error("Could not find address1.")
        return ''

    @property
    def consumption_unit(self):
        """Return the consumption."""
        return 'kWh'

    @property
    def currency(self):
        """Return the currency."""
        try:
            current_subscription = self.info['viewer']['home']['currentSubscription']
            return current_subscription['priceInfo']['current']['currency']
        except (KeyError, TypeError, IndexError):
            _LOGGER.error("Could not find currency.")
        return ''

    @property
    def country(self):
        """Return the country."""
        try:
            return self.info['viewer']['home']['address']['country']
        except (KeyError, TypeError):
            _LOGGER.error("Could not find country.")
        return ''

    @property
    def price_unit(self):
        """Return the price unit."""
        currency = self.currency
        consumption_unit = self.consumption_unit
        if not currency or not consumption_unit:
            _LOGGER.error("Could not find price_unit.")
            return ' '
        return currency + '/' + consumption_unit

    async def rt_subscribe(self, loop, async_callback):
        """Connect to Tibber and subscribe to Tibber rt subscription."""
        if self._subscription_id is not None:
            _LOGGER.error("Already subscribed.")
            return
        await self._tibber_control.rt_connect(loop)
        document = gql('''
            subscription{
              liveMeasurement(homeId:"%s"){
                timestamp
                power
                accumulatedConsumption
                accumulatedCost
                currency
                minPower
                averagePower
            maxPower
            }
           }
        ''' % self.home_id)
        sub_query = print_ast(document)

        self._subscription_id = await self._tibber_control.sub_manager.subscribe(sub_query,
                                                                                 async_callback)

    async def rt_unsubscribe(self):
        """Unsubscribe to Tibber rt subscription."""
        if self._subscription_id is None:
            _LOGGER.error("Not subscribed.")
            return
        await self._tibber_control.sub_manager.unsubscribe(self._subscription_id)

    @property
    def rt_subscription_running(self):
        """Is real time subscription running."""
        return (self._tibber_control.sub_manager is not None and
                self._tibber_control.sub_manager.is_running and
                self._subscription_id is not None
                )


class InvalidLogin(Exception):
    """Invalid login exception."""



from setuptools import setup

long_description = None
with open("README.md", 'r') as fp:
    long_description = fp.read()

setup(
    name = 'pyTibber',
    packages = ['tibber'],
    install_requires=['gql>=0.1.0', 'aiohttp>=3.0.6',
                      'async_timeout>=1.4.0', 'websockets>=6.0'],
    version='0.8.4',
    description='A python3 library to communicate with Tibber',
    long_description=long_description,
    python_requires='>=3.5.3',
    author='Daniel Hoyer Iversen',
    author_email='mail@dahoiv.net',
    url='https://github.com/Danielhiversen/pyTibber',
    license="MIT",
    classifiers=[
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Home Automation',
        'Topic :: Software Development :: Libraries :: Python Modules'
        ]
)



"""Subscription manager for Graph QL websocket."""

import asyncio
import json
import logging
from time import time

import websockets

_LOGGER = logging.getLogger(__name__)

STATE_STARTING = 'starting'
STATE_RUNNING = 'running'
STATE_STOPPED = 'stopped'


class SubscriptionManager:
    """Subscription manager."""
    # pylint: disable=too-many-instance-attributes

    def __init__(self, loop, init_payload, url):
        """Create resources for websocket communication."""
        self.loop = loop
        self.subscriptions = {}
        self._url = url
        self._state = None
        self.websocket = None
        self._retry_timer = None
        self._client_task = None
        self._wait_time_before_retry = 15
        self._session_id = 0
        self._init_payload = init_payload

    def start(self):
        """Start websocket."""
        _LOGGER.debug('Start state %s.', self._state)
        if self._state == STATE_RUNNING:
            return
        self._state = STATE_STARTING
        self._cancel_client_task()
        self._client_task = self.loop.create_task(self.running())
        for subscription_id in self.subscriptions.copy():
            callback, sub_query = self.subscriptions.pop(subscription_id, (None, None))
            _LOGGER.debug("Removed, %s", subscription_id)
            if callback is None:
                continue
            _LOGGER.debug('Add subscription %s', callback)
            self.loop.create_task(self.subscribe(sub_query, callback))

    @property
    def is_running(self):
        """Return if client is running or not."""
        return self._state == STATE_RUNNING

    async def running(self):
        """Start websocket connection."""
        await self._close_websocket()
        try:
            _LOGGER.debug("Starting")
            self.websocket = await websockets.connect(self._url,
                                                      subprotocols=["graphql-subscriptions"])
            self._state = STATE_RUNNING
            _LOGGER.debug("Running")
            self.websocket.send(json.dumps({"type": "init",
                                            "payload": self._init_payload}))

            while self._state == STATE_RUNNING:
                try:
                    msg = await asyncio.wait_for(self.websocket.recv(), timeout=30)
                except asyncio.TimeoutError:
                    _LOGGER.warning("No websocket data in 30 seconds, checking the connection.")
                    try:
                        pong_waiter = await self.websocket.ping()
                        await asyncio.wait_for(pong_waiter, timeout=10)
                    except asyncio.TimeoutError:
                        _LOGGER.error("No response to ping in 10 seconds, reconnecting.")
                        return
                    continue
                await self._process_msg(msg)
        except Exception:  # pylint: disable=broad-except
            _LOGGER.error('Unexpected error', exc_info=True)
        finally:
            await self._close_websocket()
            if self._state != STATE_STOPPED:
                _LOGGER.warning("Reconnecting")
                self._state = STATE_STOPPED
                self.retry()
            else:
                self._state = STATE_STOPPED
            _LOGGER.debug("Closing running task.")

    async def stop(self, timeout=10):
        """Close websocket connection."""
        _LOGGER.debug("Stopping client.")
        start_time = time()
        self._cancel_retry_timer()

        for subscription_id in range(len(self.subscriptions)):
            _LOGGER.debug("Sending unsubscribe: %s", subscription_id)
            await self.unsubscribe(subscription_id)

        while (timeout > 0 and
               self.websocket is not None
               and not self.subscriptions
               and (time() - start_time) < timeout/2):
            await asyncio.sleep(0.1, loop=self.loop)

        self._state = STATE_STOPPED
        while (timeout > 0 and
               self.websocket is not None and
               not self.websocket.closed and
               (time() - start_time) < timeout):
            await asyncio.sleep(0.1, loop=self.loop)

        await self._close_websocket()
        self._cancel_client_task()
        _LOGGER.debug("Server connection is stopped")

    def retry(self):
        """Retry to connect to websocket."""
        _LOGGER.debug("Retry, state: %s", self._state)
        if self._state in [STATE_STARTING, STATE_RUNNING]:
            _LOGGER.debug("Skip retry since state: %s", self._state)
            return
        self._cancel_retry_timer()
        self._state = STATE_STARTING
        self._retry_timer = self.loop.call_later(self._wait_time_before_retry, self.start)
        _LOGGER.debug('Reconnecting to server in %i seconds.', self._wait_time_before_retry)

    async def subscribe(self, sub_query, callback):
        """Add a new subscription."""

        if self.websocket is None or not self.websocket.open or not self._state == STATE_RUNNING:
            await asyncio.sleep(1, loop=self.loop)
            return await self.subscribe(sub_query, callback)

        current_session_id = self._session_id
        self._session_id += 1
        subscription = {"query": sub_query,
                        "type": "subscription_start", "id": current_session_id}
        json_subscription = json.dumps(subscription)
        await self.websocket.send(json_subscription)
        self.subscriptions[current_session_id] = (callback, sub_query)
        _LOGGER.debug("New subscription %s", current_session_id)
        return current_session_id

    async def unsubscribe(self, subscription_id):
        """Unsubscribe."""
        if self.websocket is None or not self.websocket.open:
            _LOGGER.warning("Websocket is closed.")
            return
        await self.websocket.send(json.dumps({"id": subscription_id,
                                              "type": "subscription_end"}))
        self.subscriptions.pop(subscription_id)

    async def _close_websocket(self):
        if self.websocket is None:
            return
        try:
            await self.websocket.close()
        finally:
            self.websocket = None

    async def _process_msg(self, msg):
        """Process received msg."""
        result = json.loads(msg)
        _LOGGER.debug("Recv, %s", result)

        subscription_id = result.get('id')
        if subscription_id is None:
            _LOGGER.debug('Id %s.', subscription_id)
            return

        callback, _ = self.subscriptions.get(subscription_id, (None, None))
        if callback is None:
            _LOGGER.debug('Unknown id %s.', subscription_id)
            return

        if result.get('type', '') == 'complete':
            _LOGGER.debug('Unsubscribe %s successfully.', subscription_id)
            return

        data = result.get('payload', {})
        if data is None:
            return

        await callback(data)

    def _cancel_retry_timer(self):
        if self._retry_timer:
            try:
                self._retry_timer.cancel()
            finally:
                self._retry_timer = None

    def _cancel_client_task(self):
        if self._client_task:
            try:
                self._client_task.cancel()
            finally:
                self._client_task = None
